"""DataCite for linking metadata to DOIs."""

import collections
import json
from json import JSONDecodeError
from logging import getLogger

from xmltodict import unparse

log = getLogger(__name__)

FIELD_NAME = "field_name"


def convert_datacite(metadata_record: dict, name_doi_map: dict) -> str:
    """Generate XML formatted string in DataCite format.

    Note:
        Converter is only valid for the metadata schema for EnviDat.

    Args:
        metadata_record (dict): Individual EnviDat metadata entry record dictionary.
        name_doi_map (dict): Mapping of dataset name to DOI, as dictionary name:doi.

    Returns:
        str: XML formatted string compatible with DataCite DIF 10.2 standard
    """
    try:
        converted_package = datacite_convert_dataset(metadata_record, name_doi_map)
        return unparse(converted_package, pretty=True)  # Convert OrderedDict to XML
    except ValueError as e:
        log.error(e)
        log.error("Cannot convert package to DataCite format.")
        raise ValueError("Failed to convert package to DataCite format.")


def datacite_convert_dataset(dataset: dict, name_doi_map: dict):
    """Create the DataCite XML from API dictionary."""
    # Assign datacite to ordered dictionary that will contain
    # dataset content converted to DataCite format
    datacite = collections.OrderedDict()

    # Assign language tag used several times in function
    datacite_xml_lang_tag = "xml:lang"

    # Header
    datacite["resource"] = collections.OrderedDict()
    namespace = "http://datacite.org/schema/kernel-4"
    schema = "http://schema.datacite.org/meta/kernel-4.4/metadata.xsd"
    datacite["resource"]["@xsi:schemaLocation"] = f"{namespace} {schema}"
    datacite["resource"]["@xmlns"] = f"{namespace}"
    datacite["resource"]["@xmlns:xsi"] = "http://www.w3.org/2001/XMLSchema-instance"

    # Identifier*
    datacite_identifier_tag = "identifier"
    doi = dataset.get("doi", "")
    datacite["resource"][datacite_identifier_tag] = {
        "#text": doi.strip(),
        "@identifierType": "DOI",
    }

    # Creators
    datacite_creators_tag = "creators"
    datacite_creator_tag = "creator"

    datacite["resource"][datacite_creators_tag] = {datacite_creator_tag: []}

    author_dataset = dataset.get("author", [])
    try:
        authors = json.loads(author_dataset)
    except JSONDecodeError:
        authors = []

    for author in authors:

        datacite_creator = collections.OrderedDict()

        creator_family_name = author.get("name", "").strip()
        creator_given_name = author.get("given_name", "").strip()

        if creator_given_name:
            datacite_creator[
                "creatorName"
            ] = f"{creator_given_name} {creator_family_name}"
            datacite_creator["givenName"] = creator_given_name
            datacite_creator["familyName"] = creator_family_name
        else:
            datacite_creator["creatorName"] = creator_family_name

        creator_identifier = author.get("identifier", "")
        if creator_identifier:
            datacite_creator["nameIdentifier"] = {
                "#text": creator_identifier.strip(),
                "@nameIdentifierScheme": "ORCID",
            }

        affiliations = []

        affiliation = author.get("affiliation", "")
        if affiliation:
            affiliations += [{"#text": affiliation.strip()}]

        affiliation_02 = author.get("affiliation_02", "")
        if affiliation_02:
            affiliations += [{"#text": affiliation_02.strip()}]

        affiliation_03 = author.get("affiliation_03", "")
        if affiliation_03:
            affiliations += [{"#text": affiliation_03.strip()}]

        if affiliations:
            datacite_creator["affiliation"] = affiliations

        datacite["resource"][datacite_creators_tag][datacite_creator_tag] += [
            datacite_creator
        ]

    # Titles
    datacite_titles_tag = "titles"
    datacite_title_tag = "title"
    datacite["resource"][datacite_titles_tag] = {datacite_title_tag: []}

    title = dataset.get("title", "")
    if title:
        datacite["resource"][datacite_titles_tag][datacite_title_tag] = {
            f"@{datacite_xml_lang_tag}": "en-us",
            "#text": title,
        }

    # Get publication dictionary
    pub = dataset.get("publication", {})
    try:
        publication = json.loads(pub)
    except JSONDecodeError:
        publication = {}

    # Publication year
    datacite_publication_year_tag = "publicationYear"

    publication_year = publication.get("publication_year", "")
    if publication_year:
        datacite["resource"][datacite_publication_year_tag] = {
            "#text": publication_year
        }

    # Publisher
    datacite_publisher_tag = "publisher"

    publisher = publication.get("publisher", "")
    if publisher:
        datacite["resource"][datacite_publisher_tag] = {
            f"@{datacite_xml_lang_tag}": "en-us",
            "#text": publisher.strip(),
        }

    # Subjects
    datacite_subjects = []

    # Get tags list
    tags = dataset.get("tags", [])

    for tag in tags:
        tag_name = tag.get("display_name", tag.get("name", ""))
        datacite_subjects += [{f"@{datacite_xml_lang_tag}": "en-us", "#text": tag_name}]

    if datacite_subjects:
        datacite_subjects_tag = "subjects"
        datacite_subject_tag = "subject"
        datacite["resource"][datacite_subjects_tag] = {
            datacite_subject_tag: datacite_subjects
        }

    # Contributor (contact person)
    datacite_contributors_tag = "contributors"
    datacite_contributor_tag = "contributor"

    maintainer_dataset = dataset.get("maintainer", {})
    try:
        maintainer = json.loads(maintainer_dataset)
    except JSONDecodeError:
        maintainer = {}

    datacite_contributor = collections.OrderedDict()

    contributor_family_name = maintainer.get("name", "").strip()
    contributor_given_name = maintainer.get("given_name", "").strip()

    if contributor_given_name:
        datacite_contributor[
            "contributorName"
        ] = f"{contributor_given_name} {contributor_family_name}"
        datacite_contributor["givenName"] = contributor_given_name
        datacite_contributor["familyName"] = contributor_family_name
    else:
        datacite_contributor["contributorName"] = contributor_family_name

    contributor_identifier = maintainer.get("identifier", "")
    if contributor_identifier:
        datacite_contributor["nameIdentifier"] = {
            "#text": contributor_identifier.strip(),
            "@nameIdentifierScheme": maintainer.get(
                join_tags(
                    [datacite_contributor_tag, "nameIdentifier", "nameIdentifierScheme"]
                ),
                "orcid",
            ).upper(),
        }

    contributor_affiliation = maintainer.get("affiliation", "")
    datacite_contributor["affiliation"] = contributor_affiliation.strip()

    contributor_type = maintainer.get(
        join_tags([datacite_contributor_tag, "contributorType"]), "ContactPerson"
    )
    datacite_contributor["@contributorType"] = value_to_datacite_cv(
        contributor_type, "contributorType"
    )

    if datacite_contributor:
        datacite["resource"][datacite_contributors_tag] = {
            datacite_contributor_tag: datacite_contributor
        }

    # Dates
    datacite_dates_tag = "dates"
    datacite_date_tag = "date"
    datacite_date_type_tag = "dateType"
    datacite_dates = []

    date_input = dataset.get("date", [])
    try:
        dates = json.loads(date_input)
    except JSONDecodeError:
        dates = []

    for date in dates:
        datacite_date = {
            "#text": date.get("date", ""),
            f"@{datacite_date_type_tag}": (date.get("date_type", "Valid")).title(),
        }
        datacite_dates += [datacite_date]

    if datacite_dates:
        datacite["resource"][datacite_dates_tag] = {datacite_date_tag: datacite_dates}

    # Language
    datacite_language_tag = "language"
    datacite_language = dataset.get("language", "")
    if not datacite_language:
        datacite_language = "en"
    datacite["resource"][datacite_language_tag] = {"#text": datacite_language}

    # ResourceType
    datacite_resource_type_tag = "resourceType"
    datacite_resource_type_general_tag = "resourceTypeGeneral"
    resource_type_general = dataset.get("resource_type_general", "Dataset")
    datacite_resource_type_general = value_to_datacite_cv(
        resource_type_general, datacite_resource_type_general_tag, default="Dataset"
    )

    datacite["resource"][datacite_resource_type_tag] = {
        "#text": dataset.get("resource_type", ""),
        # f'@{datacite_resource_type_general_tag}': (
        #     dataset.get('resource_type_general', 'Dataset'
        # )).title()
        f"@{datacite_resource_type_general_tag}": datacite_resource_type_general,
    }

    # Alternate Identifier
    base_url = "https://www.envidat.ch/dataset/"
    alternate_identifiers = []

    package_name = dataset.get("name", "")
    if package_name:
        package_url = f"{base_url}{package_name}"
        alternate_identifiers.append(
            {"#text": package_url, "@alternateIdentifierType": "URL"}
        )

    package_id = dataset.get("id", "")
    if package_id:
        package_id = f"{base_url}{package_id}"
        alternate_identifiers.append(
            {"#text": package_id, "@alternateIdentifierType": "URL"}
        )

    datacite["resource"]["alternateIdentifiers"] = {
        "alternateIdentifier": alternate_identifiers
    }

    # Related identifier
    related_datasets = dataset.get("related_datasets", "")
    related_datasets_base_url = "https://www.envidat.ch/#/metadata/"
    if related_datasets:

        datacite_related_urls = collections.OrderedDict()
        datacite_related_urls["relatedIdentifier"] = []

        for line in related_datasets.split("\n"):

            if line.strip().startswith("*"):
                line_contents = line.replace("*", "").strip().lower().split(" ")[0]
                related_url = None

                if line_contents in name_doi_map:
                    related_url = f"{related_datasets_base_url}{line_contents}"

                elif len(line_contents) > 0 and line_contents in name_doi_map.values():
                    related_url = f"{related_datasets_base_url}{line_contents}"

                elif line_contents.startswith("https://") or line_contents.startswith(
                    "http://"
                ):
                    related_url = line_contents

                if related_url:
                    datacite_related_urls["relatedIdentifier"] += [
                        {
                            "#text": related_url,
                            "@relatedIdentifierType": "URL",
                            "@relationType": "Cites",
                        }
                    ]

        # NOTE: Investigate including commented out block below to find
        # all URLs in 'related_datasets' key:
        # regex = (
        #     'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|'
        #     '[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+'
        # )
        # urls = re.findall(regex, related_datasets)
        #
        # for url in urls:
        #     datacite_related_urls['relatedIdentifier'] += [
        #         {
        #             '#text': url,
        #             '@relatedIdentifierType': 'URL',
        #             '@relationType': 'Cites'
        #         }
        #     ]

        if len(datacite_related_urls["relatedIdentifier"]) > 0:
            datacite["resource"]["relatedIdentifiers"] = datacite_related_urls

    # Sizes (from resources)
    datacite_size_group_tag = "sizes"
    datacite_size_tag = "size"
    datacite_sizes = []

    for resource in dataset.get("resources", []):
        if resource.get("size", ""):
            datacite_sizes += [{"#text": str(resource.get("size", " ")) + " bytes"}]
        elif resource.get("resource_size", ""):
            resource_size = resource.get("resource_size", "")
            try:
                resource_size_obj = json.loads(resource_size)
                datacite_sizes += [
                    {
                        "#text": (
                            resource_size_obj.get("size_value", "0")
                            + " "
                            + resource_size_obj.get("size_unit", "KB").upper()
                        ).strip()
                    }
                ]
            except JSONDecodeError:
                log.error("non-parsable value at resource_size:" + str(resource_size))

    if datacite_sizes:
        datacite["resource"][datacite_size_group_tag] = {
            datacite_size_tag: datacite_sizes
        }

    # Formats (from resources)
    datacite_format_group_tag = "formats"
    datacite_format_tag = "format"
    datacite_formats = []

    for resource in dataset.get("resources", []):

        default_format = resource.get("mimetype", resource.get("mimetype_inner", ""))
        resource_format = resource.get("format", "")

        if not resource_format:
            resource_format = default_format

        if resource_format:
            datacite_format = {"#text": resource_format}

            if datacite_format not in datacite_formats:
                datacite_formats += [datacite_format]

    if datacite_formats:
        datacite["resource"][datacite_format_group_tag] = {
            datacite_format_tag: datacite_formats
        }

    # Version
    datacite_version_tag = "version"
    datacite_version = dataset.get("version", "")
    if datacite_version:
        datacite["resource"][datacite_version_tag] = {"#text": datacite_version}

    # Rights
    datacite_rights_group_tag = "rightsList"
    datacite_rights_tag = "rights"
    datacite_rights_uri_tag = "rightsURI"

    datacite_scheme_uri_tag = "schemeURI"
    default_rights_scheme_uri = "https://spdx.org/licenses/"

    datacite_rights_identifier_scheme = "rightsIdentifierScheme"
    default_rights_identifier = "SPDX"

    datacite_rights_identifier = "rightsIdentifier"  # "CC0 1.0"

    rights = {}

    rights_title = dataset.get("license_title", "")
    if rights_title:
        rights = {f"@{datacite_xml_lang_tag}": "en-us", "#text": rights_title}

    rights_uri = dataset.get("license_url", "")
    if rights_uri:
        rights[f"@{datacite_rights_uri_tag}"] = rights_uri

    license_id = dataset.get("license_id", "")

    rights_id_spx = value_to_datacite_cv(
        license_id, datacite_rights_identifier, default=None
    )
    if rights_id_spx:
        rights[f"@{datacite_scheme_uri_tag}"] = default_rights_scheme_uri
        rights[f"@{datacite_rights_identifier_scheme}"] = default_rights_identifier
        rights[f"@{datacite_rights_identifier}"] = rights_id_spx

    if rights:
        datacite["resource"][datacite_rights_group_tag] = {
            datacite_rights_tag: [rights]
        }

    # Description
    datacite_descriptions_tag = "descriptions"
    datacite_description_tag = "description"
    datacite_description_type_tag = "descriptionType"
    datacite_descriptions = []

    description = dataset.get("notes", "")
    if description:
        description_text = (
            description.replace("\r", "")
            .replace(">", "-")
            .replace("<", "-")
            .replace("__", "")
            .replace("#", "")
            .replace("\n\n", "\n")
            .replace("\n\n", "\n")
        )

        datacite_description = {
            "#text": description_text.strip(),
            f"@{datacite_description_type_tag}": "Abstract",
            f"@{datacite_xml_lang_tag}": "en-us",
        }

        datacite_descriptions += [datacite_description]

    if datacite_descriptions:
        datacite["resource"][datacite_descriptions_tag] = {
            datacite_description_tag: datacite_descriptions
        }

    # GeoLocation
    datacite_geolocation_place_tag = "geoLocationPlace"

    datacite_geolocations = []
    try:
        # Get spatial data from dataset
        pkg_spatial = json.loads(dataset["spatial"])
        log.debug("pkg_spatial=" + str(pkg_spatial))
        if pkg_spatial:
            coordinates = flatten(pkg_spatial.get("coordinates", "[]"), reverse=True)
            if pkg_spatial.get("type", "").lower() == "polygon":
                datacite_geolocation = collections.OrderedDict()
                datacite_geolocation["geoLocationPolygon"] = {"polygonPoint": []}
                for coordinates_pair in pkg_spatial.get("coordinates", "[[]]")[0]:
                    geolocation_point = collections.OrderedDict()
                    geolocation_point["pointLongitude"] = coordinates_pair[0]
                    geolocation_point["pointLatitude"] = coordinates_pair[1]
                    datacite_geolocation["geoLocationPolygon"]["polygonPoint"] += [
                        geolocation_point
                    ]
                datacite_geolocations += [datacite_geolocation]
            else:
                if pkg_spatial.get("type", "").lower() == "multipoint":
                    for coordinates_pair in pkg_spatial.get("coordinates", "[]"):
                        log.debug("point=" + str(coordinates_pair))
                        datacite_geolocation = collections.OrderedDict()
                        datacite_geolocation[
                            "geoLocationPoint"
                        ] = collections.OrderedDict()
                        datacite_geolocation["geoLocationPoint"][
                            "pointLongitude"
                        ] = coordinates_pair[0]
                        datacite_geolocation["geoLocationPoint"][
                            "pointLatitude"
                        ] = coordinates_pair[1]
                        datacite_geolocations += [datacite_geolocation]
                else:
                    datacite_geolocation = collections.OrderedDict()
                    datacite_geolocation["geoLocationPoint"] = collections.OrderedDict()
                    datacite_geolocation["geoLocationPoint"][
                        "pointLongitude"
                    ] = coordinates[1]
                    datacite_geolocation["geoLocationPoint"][
                        "pointLatitude"
                    ] = coordinates[0]
                    datacite_geolocations += [datacite_geolocation]
    except JSONDecodeError:
        datacite_geolocations = []

    if datacite_geolocations:

        geolocation_place = dataset.get("spatial_info", "")
        if geolocation_place:
            datacite_geolocation_place = {
                datacite_geolocation_place_tag: geolocation_place.strip()
            }
            datacite_geolocations += [datacite_geolocation_place]

        datacite["resource"]["geoLocations"] = {"geoLocation": datacite_geolocations}

    # Funding Information
    datacite_funding_refs_tag = "fundingReferences"
    datacite_funding_ref_tag = "fundingReference"

    datacite_funding_refs = []

    funding_dataset = dataset.get("funding", [])
    try:
        funding = json.loads(funding_dataset)
    except JSONDecodeError:
        funding = []

    for funder in funding:

        datacite_funding_ref = collections.OrderedDict()

        funder_name = funder.get("institution", "")
        if funder_name:
            datacite_funding_ref["funderName"] = funder_name.strip()
            award_number = funder.get("grant_number", "")
            if award_number:
                datacite_funding_ref["awardNumber"] = award_number.strip()
            datacite_funding_refs += [datacite_funding_ref]

    if datacite_funding_refs:
        datacite["resource"][datacite_funding_refs_tag] = {
            datacite_funding_ref_tag: datacite_funding_refs
        }

    return datacite


def flatten(inp: list, reverse: bool = False) -> list:
    """Flatten list, i.e. remove a dimension/nesting."""
    output = []
    for item in inp:
        if type(item) is not list:
            if reverse:
                output = [str(item)] + output
            else:
                output += [str(item)]
        else:
            output += flatten(item, reverse)
    return output


def join_tags(tags: list, sep: str = ".") -> str:
    """Join tags by a provided separator."""
    return sep.join([tag for tag in tags if tag])


def value_to_datacite_cv(value: str, datacite_tag: str, default: str = "") -> dict:
    """Constant definitions."""
    datacite_cv = {
        "titleType": {
            "alternativetitle": "AlternativeTitle",
            "subtitle": "Subtitle",
            "translatedtitle": "TranslatedTitle",
            "other": "Other",
        },
        "resourceTypeGeneral": {
            "audiovisual": "Audiovisual",
            "collection": "Collection",
            "dataset": "Dataset",
            "event": "Event",
            "image": "Image",
            "interactiveresource": "InteractiveResource",
            "model": "Model",
            "physicalobject": "PhysicalObject",
            "service": "Service",
            "software": "Software",
            "sound": "Sound",
            "text": "Text",
            "workflow": "Workflow",
            "other": "Other",
        },
        "descriptionType": {
            "abstract": "Abstract",
            "methods": "Methods",
            "seriesinformation": "SeriesInformation",
            "tableofcontents": "TableOfContents",
            "other": "Other",
        },
        "contributorType": {
            "contactperson": "ContactPerson",
            "datacollector": "DataCollector",
            "datacurator": "DataCurator",
            "datamanager": "DataManager",
            "distributor": "Distributor",
            "editor": "Editor",
            "funder": "Funder",
            "hostinginstitution": "HostingInstitution",
            "other": "Other",
            "producer": "Producer",
            "projectleader": "ProjectLeader",
            "projectmanager": "ProjectManager",
            "projectmember": "ProjectMember",
            "registrationagency": "RegistrationAgency",
            "registrationauthority": "RegistrationAuthority",
            "relatedperson": "RelatedPerson",
            "researchgroup": "ResearchGroup",
            "rightsholder": "RightsHolder",
            "researcher": "Researcher",
            "sponsor": "Sponsor",
            "supervisor": "Supervisor",
            "workpackageleader": "WorkPackageLeader",
        },
        "rightsIdentifier": {
            "odc-odbl": "ODbL-1.0",
            "cc-by-sa": "CC-BY-SA-4.0",
            "cc-by-nc": "CC-BY-NC-4.0",
        },
    }

    # Matching ignoring blanks, case, symbols
    value_to_match = value.lower().replace(" ", "").replace("_", "")
    match_cv = datacite_cv.get(datacite_tag, {}).get(value_to_match, default)

    return match_cv


def map_fields(schema: dict, format_name: str) -> dict:
    """Map fields into correct formatting."""
    fields_map = {}
    for field in schema:
        format_field = ""
        if field.get(format_name, False):
            format_field = field[format_name]
            fields_map[format_field] = {FIELD_NAME: field[FIELD_NAME], "subfields": {}}
        for subfield in field.get("subfields", []):
            if subfield.get(format_name, False):
                format_subfield = subfield[format_name]
                if format_field:
                    if not fields_map[format_field]["subfields"].get(
                        format_subfield, False
                    ):
                        fields_map[format_field]["subfields"][format_subfield] = {
                            FIELD_NAME: subfield[FIELD_NAME]
                        }
                    else:
                        value = fields_map[format_field]["subfields"][format_subfield][
                            FIELD_NAME
                        ]
                        if isinstance(value, list):
                            fields_map[format_field]["subfields"][format_subfield] = {
                                FIELD_NAME: value + [subfield[FIELD_NAME]]
                            }
                        else:
                            fields_map[format_field]["subfields"][format_subfield] = {
                                FIELD_NAME: [value, subfield[FIELD_NAME]]
                            }
                else:
                    fields_map[format_subfield] = {
                        FIELD_NAME: field[FIELD_NAME] + "." + subfield[FIELD_NAME]
                    }
    return fields_map



"""Configuration file for PyTest tests."""

import os
from tempfile import NamedTemporaryFile
from textwrap import dedent

import pytest
from moto import mock_s3

from envidat.api.v1 import get_metadata_name_doi
from envidat.converters.bibtex_converter import bibtex_convert_dataset
from envidat.converters.datacite_converter import datacite_convert_dataset
from envidat.converters.dcat_ap_converter import dcat_ap_convert_dataset
from envidat.converters.dif_converter import dif_convert_dataset
from envidat.converters.iso_converter import iso_convert_dataset
from envidat.converters.ris_converter import ris_convert_dataset
from envidat.s3.bucket import Bucket

os.environ["MOTO_ALLOW_NONEXISTENT_REGION"] = "True"


# @pytest.fixture(scope="session")
# def s3_env_vars():

#     # Disable region validation for moto
#     os.environ["MOTO_ALLOW_NONEXISTENT_REGION"] = "True"

#     # Official vars
#     os.environ["AWS_ACCESS_KEY_ID"] = "testing"
#     os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
#     os.environ["AWS_SECURITY_TOKEN"] = "testing"
#     os.environ["AWS_SESSION_TOKEN"] = "testing"
#     os.environ["AWS_DEFAULT_REGION"] = "testing"

#     # Custom vars
#     os.environ["AWS_ENDPOINT"] = "testing"
#     os.environ["AWS_ACCESS_KEY"] = "testing"
#     os.environ["AWS_SECRET_KEY"] = "testing"
#     os.environ["AWS_REGION"] = "testing"
#     os.environ["AWS_BUCKET_NAME"] = "testing"


@pytest.fixture(scope="session")
@mock_s3
def bucket():
    """Bucket for tests."""
    Bucket.config("testing", "testing", endpoint=None, region="testing")
    new_bucket = Bucket("testing")
    return new_bucket


@pytest.fixture(scope="session")
@mock_s3
def bucket2():
    """Second bucket when two are required in tests."""
    Bucket.config("testing", "testing", endpoint=None, region="testing")
    new_bucket = Bucket("testing2")
    return new_bucket


@pytest.fixture
def create_tempfile(scope="function"):
    """Create temporary file in tests."""

    def nested_tempfile(file_type, temp_dir=None, delete=True):
        """Nested temporary file within subdirectory."""
        temp_file = NamedTemporaryFile(
            dir=temp_dir, delete=delete, suffix=f".{file_type}"
        )
        with open(temp_file.name, "w", encoding="UTF-8") as f:
            f.write("test")
        return temp_file

    return nested_tempfile


@pytest.fixture
def bibtex_converter_one_package():
    """Single package in BibTeX format."""
    package_name = "bioclim_plus"
    file_format = "bibtex"
    extension = "bib"
    return bibtex_convert_dataset, package_name, file_format, extension


@pytest.fixture
def bibtex_converter_all_packages():
    """All packages in BibTeX format."""
    file_format = "bibtex"
    extension = "bib"
    return bibtex_convert_dataset, file_format, extension


#
@pytest.fixture
def datacite_converter_one_package():
    """Single package in Datacite format."""
    package_name = (
        "ecological-properties-of-urban-ecosystems-biodiversity-dataset-of-zurich"
    )
    file_format = "datacite"
    extension = "xml"
    return (
        datacite_convert_dataset,
        get_metadata_name_doi,
        package_name,
        file_format,
        extension,
    )


@pytest.fixture
def datacite_converter_all_packages():
    """All packages in Datacite format."""
    file_format = "datacite"
    extension = "xml"
    return datacite_convert_dataset, get_metadata_name_doi, file_format, extension


@pytest.fixture
def dif_converter_one_package():
    """Single package in Diff format."""
    package_name = "resolution-in-sdms-shapes-plant-multifaceted-diversity"
    file_format = "gcmd_dif"
    extension = "xml"
    return dif_convert_dataset, package_name, file_format, extension


@pytest.fixture
def dif_converter_all_packages():
    """All packages in Diff format."""
    file_format = "gcmd_dif"
    extension = "xml"
    return dif_convert_dataset, file_format, extension


@pytest.fixture
def iso_converter_one_package():
    """Single package in ISO format."""
    package_name = "intratrait"
    file_format = "iso19139"
    extension = "xml"
    return iso_convert_dataset, package_name, file_format, extension


@pytest.fixture
def iso_converter_all_packages():
    """All packages in ISO format."""
    file_format = "iso19139"
    extension = "xml"
    return iso_convert_dataset, file_format, extension


@pytest.fixture
def dcat_ap_converter_all_packages():
    """All packages in DCAT-AP format."""
    file_format = "dcat-ap-ch"
    extension = "xml"
    return dcat_ap_convert_dataset, file_format, extension


@pytest.fixture
def ris_converter_one_package():
    """Single package in RIS format."""
    package_name = "bioclim_plus"
    file_format = "ris"
    extension = "ris"
    return ris_convert_dataset, package_name, file_format, extension


@pytest.fixture
def ris_converter_all_packages():
    """All packages in RIS format."""
    file_format = "ris"
    extension = "ris"
    return ris_convert_dataset, file_format, extension


@pytest.fixture
def metadata_keys():
    """List of keys required for an EnviDat metadata Record."""
    return [
        "author",
        "author_email",
        "creator_user_id",
        "date",
        "doi",
        "funding",
        "id",
        "isopen",
        "language",
        "license_id",
        "license_title",
        "maintainer",
        "maintainer_email",
        "metadata_created",
        "metadata_modified",
        "name",
        "notes",
        "num_resources",
        "num_tags",
        "organization",
        "owner_org",
        "private",
        "publication",
        "publication_state",
        "related_datasets",
        "related_publications",
        "resource_type",
        "resource_type_general",
        "spatial",
        "spatial_info",
        "state",
        "subtitle",
        "title",
        "type",
        "url",
        "version",
        "resources",
        "tags",
        "groups",
        "relationships_as_subject",
        "relationships_as_object",
    ]


@pytest.fixture
def example_ckan_dict():
    """CKAN metadata dict example for use in tests."""
    return {
        "author": '[{"name": "Casanelles-Abella", "data_credit": ["validation", "curation", "software", "publication"], "affiliation": "Conservation Biology, Biodiversity and Conservation Biology, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland", "affiliation_03": "", "affiliation_02": "Landscape Ecology, Institute of Terrestrial Ecosystems, ETH Z\\u00fcrich, 8049 Z\\u00fcrich, Switzerland", "identifier": "0000-0003-1924-9298", "email": "joan.casanelles@wsl.ch", "given_name": "Joan"}, {"name": "Chauvier", "data_credit": ["software", "publication"], "email": "yohann.chauvier@wsl.ch", "affiliation": "Dynamic Macroecology, Landscape Change Science, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland", "affiliation_03": "", "given_name": "Yohann", "identifier": "0000-0001-9399-3192", "affiliation_02": ""}, {"name": "Zellweger", "data_credit": ["software", "publication"], "email": "florian.zellweger@wsl.ch", "affiliation": "Resource Analysis, Forest Resource and Management, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland", "affiliation_03": "", "affiliation_02": "", "identifier": "0000-0003-1265-9147", "given_name": "Florian"}, {"name": "Villiger", "data_credit": ["software", "publication"], "affiliation": "Landscape Ecology, Institute of Terrestrial Ecosystems, ETH Z\\u00fcrich, 8049 Z\\u00fcrich, Switzerland", "affiliation_03": "", "given_name": "Petrissa", "identifier": "", "email": "villigep@student.ethz.ch", "affiliation_02": ""}, {"name": "Frey", "data_credit": ["software", "publication"], "affiliation": "Conservation Biology, Biodiversity and Conservation Biology, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland", "affiliation_03": "", "affiliation_02": "Spatial Evolutionary Ecology, Biodiversity and Conservation Biology, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland", "identifier": "0000-0002-4603-0438", "email": "david.frey@wsl.ch", "given_name": "David Johannes"}, {"name": "Ginzler", "data_credit": ["software", "publication"], "affiliation": "Remote Sensing, Landscape Change Science, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland", "affiliation_03": "", "affiliation_02": "", "identifier": "0000-0001-6365-2151", "email": "christian.ginzler@wsl.ch", "given_name": "Christian"}, {"name": "Moretti", "data_credit": ["collection", "publication", "supervision"], "email": "marco.moretti@wsl.ch", "affiliation": "Conservation Biology, Biodiversity and Conservation Biology, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland", "affiliation_03": "", "given_name": "Marco", "identifier": "0000-0002-5845-3198", "affiliation_02": ""}, {"name": "Pellissier", "data_credit": ["publication", "supervision"], "affiliation": "Landscape Ecology, Institute of Terrestrial Ecosystems, ETH Z\\u00fcrich, 8049 Z\\u00fcrich, Switzerland", "affiliation_03": "", "given_name": "Loic", "identifier": "0000-0002-2289-8259", "email": "loic.pellissier@wsl.ch", "affiliation_02": "Landscape Ecology, Landscape Change Science, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland"}]',
        "author_email": None,
        "creator_user_id": "c7314170-c2ac-46d6-9193-eee220229e8b",
        "date": '[{"date": "01.08.2006", "date_type": "collected", "end_date": "16.09.2007"}, {"date": "25.05.2010", "date_type": "collected", "end_date": "03.09.2010"}, {"date": "06.01.2015", "date_type": "collected", "end_date": "20.09.2016"}, {"date": "2017-02-01", "date_type": "collected", "end_date": "2017-10-31"}]',
        "doi": "10.16904/envidat.172",
        "funding": '[{"grant_number": "H2020, BiodivERsA32015104", "institution": "Swiss National Science Foundation (project 31BD30_172467) within the programme ERA-Net BiodivERsA project “BioVEINS: Connectivity of green and blue infrastructures: living veins for biodiverse and healthy cities” (H2020 BiodivERsA32015104)", "institution_url": "http://bioveins.eu, https://www.biodiversa.org, http://p3.snf.ch/project-172467"}, {"grant_number": "172198", "institution": "F.Z. received funding from the Swiss National Science Foundation (project 172198)", "institution_url": "http://p3.snf.ch/project-172198"}, {"grant_number": "310030L_170059", "institution": "Y.C. acknowledges the ANR-SNF bilateral project OriginAlps (grant number 310030L_170059)", "institution_url": "http://p3.snf.ch/project-170059"}]',
        "id": "0f81429a-204e-4720-929e-5b251e9ab1c4",
        "isopen": False,
        "language": "en",
        "license_id": "other-undefined",
        "license_title": "Other (Specified in the description)",
        "maintainer": '{"affiliation": "Conservation Biology, Biodiversity and Conservation Biology, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland", "email": "marco.moretti@wsl.ch", "name": "Moretti", "identifier": "0000-0002-5845-3198", "given_name": "Marco"}',
        "maintainer_email": None,
        "metadata_created": "2020-09-07T15:14:02.051289",
        "metadata_modified": "2020-09-17T09:14:48.641088",
        "name": "ecological-properties-of-urban-ecosystems-biodiversity-dataset-of-zurich",
        "notes": "Richness, site occurrence and abundance data of bees, beetles, birds, hoverflies, net-wingeds, true bugs, snails, spiders, milipides, wasps collected in the city of Zurich using different sampling techniques, and the environmental variables for each sampling site.\r\nData are provided on request to contact person against bilateral agreement.",
        "num_resources": 5,
        "num_tags": 11,
        "organization": {
            "id": "32672ec5-c180-471e-92f7-ef58e00eb1fb",
            "name": "conservation-biology",
            "title": "Conservation Biology",
            "type": "organization",
            "description": "We study the relationship between threatened species and their habitat with the objective to develop evidence-based instruments for conservation. The viability of populations of habitat specialists is often limited due to restricted availability of suitable habitat or discontinuous distribution. We focus on selected animal species of conservation concern (forest birds, bats, forest insects) and analyse species-habitat relationships at a scale ranging from single subpopulations to entire metapopulations. We strive towards standardised recording of elusive species with novel techniques, assessing environmental variables and combining these field data with data from remote sensing. We determine limiting factors (habitat, climate, resources) and predict species occurrence and habitat suitability with modelling techniques.\r\n\r\nThese models allow to predict species occurrence patterns and to delineate evidence-based priority areas for conservation. Our long-term collaboration with federal and cantonal stakeholders of forest and nature conservation enables the combination of research results with conservation requirements (e.g. species action plans, forest biodiversity expert panel, scientific counsel for bat protection). While we investigate complete distribution ranges in bats, we focus on forest habitats in birds and insects given the large extent and high importance of forests for biodiversity conservation in Switzerland.",
            "image_url": "2018-07-10-090115.162284LogoWSL.svg",
            "created": "2018-05-18T09:12:30.577470",
            "is_organization": True,
            "approval_status": "approved",
            "state": "active",
        },
        "owner_org": "32672ec5-c180-471e-92f7-ef58e00eb1fb",
        "private": False,
        "publication": '{"publisher": "EnviDat", "publication_year": "2020"}',
        "publication_state": "published",
        "related_datasets": "* Dataset_1_SAD_SOD.txt\r\n* Dataset_2_SDM_data.txt\r\n* Dataset_3_Species_richness_hyper.txt\r\n* Dataset_4_Species_richness_rare.txt\r\n* Dataset_5_Species_richness_all.txt\r\n*\r\nData extracts in TAB-delimted format. First row is header",
        "related_publications": "",
        "resource_type": "dataset",
        "resource_type_general": "dataset",
        "spatial": '{"type":"Polygon","coordinates":[[[8.463935852050781,47.329748306633896],[8.463935852050781,47.427622733649606],[8.602638244628906,47.427622733649606],[8.602638244628906,47.329748306633896],[8.463935852050781,47.329748306633896]]]}',
        "spatial_info": "Switzerland",
        "state": "active",
        "subtitle": "",
        "title": "Ecological properties of urban ecosystems. Biodiversity dataset of Zurich",
        "type": "dataset",
        "url": None,
        "version": "1.0",
        "resources": [
            {
                "cache_last_updated": None,
                "cache_url": None,
                "created": "2020-09-07T15:18:31.256479",
                "description": "abundance and occurrence data of the 1446 species studied, used to investigate species-abundance distributions and species-occurrence distributions",
                "doi": "",
                "format": "TXT",
                "hash": "",
                "id": "a7a96a85-dc46-4b5f-bcf4-4d91485e3d03",
                "last_modified": "2020-09-07T15:18:31.071900",
                "metadata_modified": None,
                "mimetype": "text/plain",
                "mimetype_inner": None,
                "name": "Dataset_1_SAD_SOD.txt",
                "package_id": "0f81429a-204e-4720-929e-5b251e9ab1c4",
                "position": 0,
                "resource_size": '{"size_value": "", "size_units": "kb"}',
                "resource_type": None,
                "restricted": '{"shared_secret": "", "allowed_users": "marco_moretti-wsl_ch,joan_casanelles-wsl_ch", "level": "same_organization"}',
                "size": 73478,
                "state": "active",
                "url": "https://www.envidat.ch/dataset/0f81429a-204e-4720-929e-5b251e9ab1c4/resource/a7a96a85-dc46-4b5f-bcf4-4d91485e3d03/download/dataset_1_sad_sod.txt",
                "url_type": "upload",
            },
            {
                "cache_last_updated": None,
                "cache_url": None,
                "created": "2020-09-07T15:19:36.603618",
                "description": "contains the raw data to perform species distribution models, only for common species",
                "doi": "",
                "format": "TXT",
                "hash": "",
                "id": "0e2e7fb0-88ba-4aa4-92ef-079191b4a668",
                "last_modified": "2020-09-07T15:19:36.419018",
                "metadata_modified": None,
                "mimetype": "text/plain",
                "mimetype_inner": None,
                "name": "Dataset_2_SDM_data.txt",
                "package_id": "0f81429a-204e-4720-929e-5b251e9ab1c4",
                "position": 1,
                "resource_size": '{"size_value": "", "size_units": "kb"}',
                "resource_type": None,
                "restricted": '{"level": "same_organization", "allowed_users": "marco_moretti-wsl_ch,joan_casanelles-wsl_ch", "shared_secret": ""}',
                "size": 65679611,
                "state": "active",
                "url": "https://www.envidat.ch/dataset/0f81429a-204e-4720-929e-5b251e9ab1c4/resource/0e2e7fb0-88ba-4aa4-92ef-079191b4a668/download/dataset_2_sdm_data.txt",
                "url_type": "upload",
            },
            {
                "cache_last_updated": None,
                "cache_url": None,
                "created": "2020-09-07T15:20:43.446679",
                "description": "contains the raw data to perform the species richness models of the hypercommon species",
                "doi": "",
                "format": "TXT",
                "hash": "",
                "id": "0ef3b508-b8e7-4404-a7b9-edc51c7f2375",
                "last_modified": "2020-09-07T15:20:43.249061",
                "metadata_modified": None,
                "mimetype": "text/plain",
                "mimetype_inner": None,
                "name": "Dataset_3_Species_richness_hyper.txt",
                "package_id": "0f81429a-204e-4720-929e-5b251e9ab1c4",
                "position": 2,
                "resource_size": '{"size_value": "", "size_units": "kb"}',
                "resource_type": None,
                "restricted": '{"shared_secret": "", "allowed_users": "marco_moretti-wsl_ch,joan_casanelles-wsl_ch", "level": "same_organization"}',
                "size": 1373081,
                "state": "active",
                "url": "https://www.envidat.ch/dataset/0f81429a-204e-4720-929e-5b251e9ab1c4/resource/0ef3b508-b8e7-4404-a7b9-edc51c7f2375/download/dataset_3_species_richness_hyper.txt",
                "url_type": "upload",
            },
            {
                "cache_last_updated": None,
                "cache_url": None,
                "created": "2020-09-07T15:21:27.435458",
                "description": "contains the raw data to perform the species richness models of the rare species",
                "doi": "",
                "format": "TXT",
                "hash": "",
                "id": "3bbbc78a-92c3-4832-b51b-61d65346460e",
                "last_modified": "2020-09-07T15:21:27.252544",
                "metadata_modified": None,
                "mimetype": "text/plain",
                "mimetype_inner": None,
                "name": "Dataset_4_Species_richness_rare.txt",
                "package_id": "0f81429a-204e-4720-929e-5b251e9ab1c4",
                "position": 3,
                "resource_size": '{"size_value": "", "size_units": "kb"}',
                "resource_type": None,
                "restricted": '{"level": "same_organization", "allowed_users": "marco_moretti-wsl_ch,joan_casanelles-wsl_ch", "shared_secret": ""}',
                "size": 2274863,
                "state": "active",
                "url": "https://www.envidat.ch/dataset/0f81429a-204e-4720-929e-5b251e9ab1c4/resource/3bbbc78a-92c3-4832-b51b-61d65346460e/download/dataset_4_species_richness_rare.txt",
                "url_type": "upload",
            },
            {
                "cache_last_updated": None,
                "cache_url": None,
                "created": "2020-09-07T15:22:28.765321",
                "description": "contains the raw data to perform the species richness models of all species\r\n",
                "doi": "",
                "format": "TXT",
                "hash": "",
                "id": "f3abc8d4-c92e-4a57-aea4-39ba8c93158d",
                "last_modified": "2020-09-07T15:22:28.580170",
                "metadata_modified": None,
                "mimetype": "text/plain",
                "mimetype_inner": None,
                "name": "Dataset_5_Species_richness_all.txt",
                "package_id": "0f81429a-204e-4720-929e-5b251e9ab1c4",
                "position": 4,
                "resource_size": '{"size_value": "", "size_units": "kb"}',
                "resource_type": None,
                "restricted": '{"shared_secret": "", "allowed_users": "marco_moretti-wsl_ch,joan_casanelles-wsl_ch", "level": "same_organization"}',
                "size": 2452421,
                "state": "active",
                "url": "https://www.envidat.ch/dataset/0f81429a-204e-4720-929e-5b251e9ab1c4/resource/f3abc8d4-c92e-4a57-aea4-39ba8c93158d/download/dataset_5_species_richness_all.txt",
                "url_type": "upload",
            },
        ],
        "tags": [
            {
                "display_name": "ARTHROPODS",
                "id": "933b2f75-a0ed-4a25-a984-44b66f0c3bfd",
                "name": "ARTHROPODS",
                "state": "active",
                "vocabulary_id": None,
            },
            {
                "display_name": "BIRDS",
                "id": "0e8e27f0-fc1f-49f3-bb09-717cbccb6c32",
                "name": "BIRDS",
                "state": "active",
                "vocabulary_id": None,
            },
            {
                "display_name": "DIVERSITY",
                "id": "6333fb3f-2838-4496-88bb-3e67160e2d4d",
                "name": "DIVERSITY",
                "state": "active",
                "vocabulary_id": None,
            },
            {
                "display_name": "INSECTS",
                "id": "727d285b-f4e8-4bc7-8e71-e63ed9252bf6",
                "name": "INSECTS",
                "state": "active",
                "vocabulary_id": None,
            },
            {
                "display_name": "SDMS",
                "id": "1feba39d-6f50-4300-b5ba-6adc00490ee0",
                "name": "SDMS",
                "state": "active",
                "vocabulary_id": None,
            },
            {
                "display_name": "SNAILS",
                "id": "92a892d7-13e7-43a7-98b2-339c6cb4b3c1",
                "name": "SNAILS",
                "state": "active",
                "vocabulary_id": None,
            },
            {
                "display_name": "SPECIES ABUNDANCE DISTRIBUTIONS",
                "id": "3d02d873-d64b-4164-b73a-d0123c3e7128",
                "name": "SPECIES ABUNDANCE DISTRIBUTIONS",
                "state": "active",
                "vocabulary_id": None,
            },
            {
                "display_name": "SPECIES RICHNESS MODELS",
                "id": "e61b9a3f-fa0d-4e90-ae32-279ad9ceda29",
                "name": "SPECIES RICHNESS MODELS",
                "state": "active",
                "vocabulary_id": None,
            },
            {
                "display_name": "SPIDERS",
                "id": "3762a970-222d-4cb3-ab58-bf79e164e5ed",
                "name": "SPIDERS",
                "state": "active",
                "vocabulary_id": None,
            },
            {
                "display_name": "URBAN ECOLOGY",
                "id": "68337815-ea2a-4f12-8c28-013f9971878e",
                "name": "URBAN ECOLOGY",
                "state": "active",
                "vocabulary_id": None,
            },
            {
                "display_name": "URBANIZATION",
                "id": "80141c83-e3e9-43b5-9c76-abad1e74852c",
                "name": "URBANIZATION",
                "state": "active",
                "vocabulary_id": None,
            },
        ],
        "groups": [],
        "relationships_as_subject": [],
        "relationships_as_object": [],
    }


@pytest.fixture
def example_ckan_json():
    """CKAN metadata JSON example for use in tests."""
    return dedent(
        r"""
        {
            "author": "[{\"name\": \"Casanelles-Abella\", \"data_credit\": [\"validation\", \"curation\", \"software\", \"publication\"], \"affiliation\": \"Conservation Biology, Biodiversity and Conservation Biology, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland\", \"affiliation_03\": \"\", \"affiliation_02\": \"Landscape Ecology, Institute of Terrestrial Ecosystems, ETH Z\\u00fcrich, 8049 Z\\u00fcrich, Switzerland\", \"identifier\": \"0000-0003-1924-9298\", \"email\": \"joan.casanelles@wsl.ch\", \"given_name\": \"Joan\"}, {\"name\": \"Chauvier\", \"data_credit\": [\"software\", \"publication\"], \"email\": \"yohann.chauvier@wsl.ch\", \"affiliation\": \"Dynamic Macroecology, Landscape Change Science, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland\", \"affiliation_03\": \"\", \"given_name\": \"Yohann\", \"identifier\": \"0000-0001-9399-3192\", \"affiliation_02\": \"\"}, {\"name\": \"Zellweger\", \"data_credit\": [\"software\", \"publication\"], \"email\": \"florian.zellweger@wsl.ch\", \"affiliation\": \"Resource Analysis, Forest Resource and Management, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland\", \"affiliation_03\": \"\", \"affiliation_02\": \"\", \"identifier\": \"0000-0003-1265-9147\", \"given_name\": \"Florian\"}, {\"name\": \"Villiger\", \"data_credit\": [\"software\", \"publication\"], \"affiliation\": \"Landscape Ecology, Institute of Terrestrial Ecosystems, ETH Z\\u00fcrich, 8049 Z\\u00fcrich, Switzerland\", \"affiliation_03\": \"\", \"given_name\": \"Petrissa\", \"identifier\": \"\", \"email\": \"villigep@student.ethz.ch\", \"affiliation_02\": \"\"}, {\"name\": \"Frey\", \"data_credit\": [\"software\", \"publication\"], \"affiliation\": \"Conservation Biology, Biodiversity and Conservation Biology, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland\", \"affiliation_03\": \"\", \"affiliation_02\": \"Spatial Evolutionary Ecology, Biodiversity and Conservation Biology, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland\", \"identifier\": \"0000-0002-4603-0438\", \"email\": \"david.frey@wsl.ch\", \"given_name\": \"David Johannes\"}, {\"name\": \"Ginzler\", \"data_credit\": [\"software\", \"publication\"], \"affiliation\": \"Remote Sensing, Landscape Change Science, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland\", \"affiliation_03\": \"\", \"affiliation_02\": \"\", \"identifier\": \"0000-0001-6365-2151\", \"email\": \"christian.ginzler@wsl.ch\", \"given_name\": \"Christian\"}, {\"name\": \"Moretti\", \"data_credit\": [\"collection\", \"publication\", \"supervision\"], \"email\": \"marco.moretti@wsl.ch\", \"affiliation\": \"Conservation Biology, Biodiversity and Conservation Biology, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland\", \"affiliation_03\": \"\", \"given_name\": \"Marco\", \"identifier\": \"0000-0002-5845-3198\", \"affiliation_02\": \"\"}, {\"name\": \"Pellissier\", \"data_credit\": [\"publication\", \"supervision\"], \"affiliation\": \"Landscape Ecology, Institute of Terrestrial Ecosystems, ETH Z\\u00fcrich, 8049 Z\\u00fcrich, Switzerland\", \"affiliation_03\": \"\", \"given_name\": \"Loic\", \"identifier\": \"0000-0002-2289-8259\", \"email\": \"loic.pellissier@wsl.ch\", \"affiliation_02\": \"Landscape Ecology, Landscape Change Science, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland\"}]",
            "author_email": null,
            "creator_user_id": "c7314170-c2ac-46d6-9193-eee220229e8b",
            "date": "[{\"date\": \"01.08.2006\", \"date_type\": \"collected\", \"end_date\": \"16.09.2007\"}, {\"date\": \"25.05.2010\", \"date_type\": \"collected\", \"end_date\": \"03.09.2010\"}, {\"date\": \"06.01.2015\", \"date_type\": \"collected\", \"end_date\": \"20.09.2016\"}, {\"date\": \"2017-02-01\", \"date_type\": \"collected\", \"end_date\": \"2017-10-31\"}]",
            "doi": "10.16904/envidat.172",
            "funding": "[{\"grant_number\": \"H2020, BiodivERsA32015104\", \"institution\": \"Swiss National Science Foundation (project 31BD30_172467) within the programme ERA-Net BiodivERsA project \u201cBioVEINS: Connectivity of green and blue infrastructures: living veins for biodiverse and healthy cities\u201d (H2020 BiodivERsA32015104)\", \"institution_url\": \"http://bioveins.eu, https://www.biodiversa.org, http://p3.snf.ch/project-172467\"}, {\"grant_number\": \"172198\", \"institution\": \"F.Z. received funding from the Swiss National Science Foundation (project 172198)\", \"institution_url\": \"http://p3.snf.ch/project-172198\"}, {\"grant_number\": \"310030L_170059\", \"institution\": \"Y.C. acknowledges the ANR-SNF bilateral project OriginAlps (grant number 310030L_170059)\", \"institution_url\": \"http://p3.snf.ch/project-170059\"}]",
            "id": "0f81429a-204e-4720-929e-5b251e9ab1c4",
            "isopen": false,
            "language": "en",
            "license_id": "other-undefined",
            "license_title": "Other (Specified in the description)",
            "maintainer": "{\"affiliation\": \"Conservation Biology, Biodiversity and Conservation Biology, Swiss Federal Institute for Forest, Snow and Landscape Research WSL, 8903 Birmensdorf, Switzerland\", \"email\": \"marco.moretti@wsl.ch\", \"name\": \"Moretti\", \"identifier\": \"0000-0002-5845-3198\", \"given_name\": \"Marco\"}",
            "maintainer_email": null,
            "metadata_created": "2020-09-07T15:14:02.051289",
            "metadata_modified": "2020-09-17T09:14:48.641088",
            "name": "ecological-properties-of-urban-ecosystems-biodiversity-dataset-of-zurich",
            "notes": "Richness, site occurrence and abundance data of bees, beetles, birds, hoverflies, net-wingeds, true bugs, snails, spiders, milipides, wasps collected in the city of Zurich using different sampling techniques, and the environmental variables for each sampling site.\r\nData are provided on request to contact person against bilateral agreement.",
            "num_resources": 5,
            "num_tags": 11,
            "organization": {
                "id": "32672ec5-c180-471e-92f7-ef58e00eb1fb",
                "name": "conservation-biology",
                "title": "Conservation Biology",
                "type": "organization",
                "description": "We study the relationship between threatened species and their habitat with the objective to develop evidence-based instruments for conservation. The viability of populations of habitat specialists is often limited due to restricted availability of suitable habitat or discontinuous distribution. We focus on selected animal species of conservation concern (forest birds, bats, forest insects) and analyse species-habitat relationships at a scale ranging from single subpopulations to entire metapopulations. We strive towards standardised recording of elusive species with novel techniques, assessing environmental variables and combining these field data with data from remote sensing. We determine limiting factors (habitat, climate, resources) and predict species occurrence and habitat suitability with modelling techniques.\r\n\r\nThese models allow to predict species occurrence patterns and to delineate evidence-based priority areas for conservation. Our long-term collaboration with federal and cantonal stakeholders of forest and nature conservation enables the combination of research results with conservation requirements (e.g. species action plans, forest biodiversity expert panel, scientific counsel for bat protection). While we investigate complete distribution ranges in bats, we focus on forest habitats in birds and insects given the large extent and high importance of forests for biodiversity conservation in Switzerland.",
                "image_url": "2018-07-10-090115.162284LogoWSL.svg",
                "created": "2018-05-18T09:12:30.577470",
                "is_organization": true,
                "approval_status": "approved",
                "state": "active"
            },
            "owner_org": "32672ec5-c180-471e-92f7-ef58e00eb1fb",
            "private": false,
            "publication": "{\"publisher\": \"EnviDat\", \"publication_year\": \"2020\"}",
            "publication_state": "published",
            "related_datasets": "* Dataset_1_SAD_SOD.txt\r\n* Dataset_2_SDM_data.txt\r\n* Dataset_3_Species_richness_hyper.txt\r\n* Dataset_4_Species_richness_rare.txt\r\n* Dataset_5_Species_richness_all.txt\r\n*\r\nData extracts in TAB-delimted format. First row is header",
            "related_publications": "",
            "resource_type": "dataset",
            "resource_type_general": "dataset",
            "spatial": "{\"type\":\"Polygon\",\"coordinates\":[[[8.463935852050781,47.329748306633896],[8.463935852050781,47.427622733649606],[8.602638244628906,47.427622733649606],[8.602638244628906,47.329748306633896],[8.463935852050781,47.329748306633896]]]}",
            "spatial_info": "Switzerland",
            "state": "active",
            "subtitle": "",
            "title": "Ecological properties of urban ecosystems. Biodiversity dataset of Zurich",
            "type": "dataset",
            "url": null,
            "version": "1.0",
            "resources": [
                {
                    "cache_last_updated": null,
                    "cache_url": null,
                    "created": "2020-09-07T15:18:31.256479",
                    "description": "abundance and occurrence data of the 1446 species studied, used to investigate species-abundance distributions and species-occurrence distributions",
                    "doi": "",
                    "format": "TXT",
                    "hash": "",
                    "id": "a7a96a85-dc46-4b5f-bcf4-4d91485e3d03",
                    "last_modified": "2020-09-07T15:18:31.071900",
                    "metadata_modified": null,
                    "mimetype": "text/plain",
                    "mimetype_inner": null,
                    "name": "Dataset_1_SAD_SOD.txt",
                    "package_id": "0f81429a-204e-4720-929e-5b251e9ab1c4",
                    "position": 0,
                    "resource_size": "{\"size_value\": \"\", \"size_units\": \"kb\"}",
                    "resource_type": null,
                    "restricted": "{\"shared_secret\": \"\", \"allowed_users\": \"marco_moretti-wsl_ch,joan_casanelles-wsl_ch\", \"level\": \"same_organization\"}",
                    "size": 73478,
                    "state": "active",
                    "url": "https://www.envidat.ch/dataset/0f81429a-204e-4720-929e-5b251e9ab1c4/resource/a7a96a85-dc46-4b5f-bcf4-4d91485e3d03/download/dataset_1_sad_sod.txt",
                    "url_type": "upload"
                },
                {
                    "cache_last_updated": null,
                    "cache_url": null,
                    "created": "2020-09-07T15:19:36.603618",
                    "description": "contains the raw data to perform species distribution models, only for common species",
                    "doi": "",
                    "format": "TXT",
                    "hash": "",
                    "id": "0e2e7fb0-88ba-4aa4-92ef-079191b4a668",
                    "last_modified": "2020-09-07T15:19:36.419018",
                    "metadata_modified": null,
                    "mimetype": "text/plain",
                    "mimetype_inner": null,
                    "name": "Dataset_2_SDM_data.txt",
                    "package_id": "0f81429a-204e-4720-929e-5b251e9ab1c4",
                    "position": 1,
                    "resource_size": "{\"size_value\": \"\", \"size_units\": \"kb\"}",
                    "resource_type": null,
                    "restricted": "{\"level\": \"same_organization\", \"allowed_users\": \"marco_moretti-wsl_ch,joan_casanelles-wsl_ch\", \"shared_secret\": \"\"}",
                    "size": 65679611,
                    "state": "active",
                    "url": "https://www.envidat.ch/dataset/0f81429a-204e-4720-929e-5b251e9ab1c4/resource/0e2e7fb0-88ba-4aa4-92ef-079191b4a668/download/dataset_2_sdm_data.txt",
                    "url_type": "upload"
                },
                {
                    "cache_last_updated": null,
                    "cache_url": null,
                    "created": "2020-09-07T15:20:43.446679",
                    "description": "contains the raw data to perform the species richness models of the hypercommon species",
                    "doi": "",
                    "format": "TXT",
                    "hash": "",
                    "id": "0ef3b508-b8e7-4404-a7b9-edc51c7f2375",
                    "last_modified": "2020-09-07T15:20:43.249061",
                    "metadata_modified": null,
                    "mimetype": "text/plain",
                    "mimetype_inner": null,
                    "name": "Dataset_3_Species_richness_hyper.txt",
                    "package_id": "0f81429a-204e-4720-929e-5b251e9ab1c4",
                    "position": 2,
                    "resource_size": "{\"size_value\": \"\", \"size_units\": \"kb\"}",
                    "resource_type": null,
                    "restricted": "{\"shared_secret\": \"\", \"allowed_users\": \"marco_moretti-wsl_ch,joan_casanelles-wsl_ch\", \"level\": \"same_organization\"}",
                    "size": 1373081,
                    "state": "active",
                    "url": "https://www.envidat.ch/dataset/0f81429a-204e-4720-929e-5b251e9ab1c4/resource/0ef3b508-b8e7-4404-a7b9-edc51c7f2375/download/dataset_3_species_richness_hyper.txt",
                    "url_type": "upload"
                },
                {
                    "cache_last_updated": null,
                    "cache_url": null,
                    "created": "2020-09-07T15:21:27.435458",
                    "description": "contains the raw data to perform the species richness models of the rare species",
                    "doi": "",
                    "format": "TXT",
                    "hash": "",
                    "id": "3bbbc78a-92c3-4832-b51b-61d65346460e",
                    "last_modified": "2020-09-07T15:21:27.252544",
                    "metadata_modified": null,
                    "mimetype": "text/plain",
                    "mimetype_inner": null,
                    "name": "Dataset_4_Species_richness_rare.txt",
                    "package_id": "0f81429a-204e-4720-929e-5b251e9ab1c4",
                    "position": 3,
                    "resource_size": "{\"size_value\": \"\", \"size_units\": \"kb\"}",
                    "resource_type": null,
                    "restricted": "{\"level\": \"same_organization\", \"allowed_users\": \"marco_moretti-wsl_ch,joan_casanelles-wsl_ch\", \"shared_secret\": \"\"}",
                    "size": 2274863,
                    "state": "active",
                    "url": "https://www.envidat.ch/dataset/0f81429a-204e-4720-929e-5b251e9ab1c4/resource/3bbbc78a-92c3-4832-b51b-61d65346460e/download/dataset_4_species_richness_rare.txt",
                    "url_type": "upload"
                },
                {
                    "cache_last_updated": null,
                    "cache_url": null,
                    "created": "2020-09-07T15:22:28.765321",
                    "description": "contains the raw data to perform the species richness models of all species\r\n",
                    "doi": "",
                    "format": "TXT",
                    "hash": "",
                    "id": "f3abc8d4-c92e-4a57-aea4-39ba8c93158d",
                    "last_modified": "2020-09-07T15:22:28.580170",
                    "metadata_modified": null,
                    "mimetype": "text/plain",
                    "mimetype_inner": null,
                    "name": "Dataset_5_Species_richness_all.txt",
                    "package_id": "0f81429a-204e-4720-929e-5b251e9ab1c4",
                    "position": 4,
                    "resource_size": "{\"size_value\": \"\", \"size_units\": \"kb\"}",
                    "resource_type": null,
                    "restricted": "{\"shared_secret\": \"\", \"allowed_users\": \"marco_moretti-wsl_ch,joan_casanelles-wsl_ch\", \"level\": \"same_organization\"}",
                    "size": 2452421,
                    "state": "active",
                    "url": "https://www.envidat.ch/dataset/0f81429a-204e-4720-929e-5b251e9ab1c4/resource/f3abc8d4-c92e-4a57-aea4-39ba8c93158d/download/dataset_5_species_richness_all.txt",
                    "url_type": "upload"
                }
            ],
            "tags": [
                {
                    "display_name": "ARTHROPODS",
                    "id": "933b2f75-a0ed-4a25-a984-44b66f0c3bfd",
                    "name": "ARTHROPODS",
                    "state": "active",
                    "vocabulary_id": null
                },
                {
                    "display_name": "BIRDS",
                    "id": "0e8e27f0-fc1f-49f3-bb09-717cbccb6c32",
                    "name": "BIRDS",
                    "state": "active",
                    "vocabulary_id": null
                },
                {
                    "display_name": "DIVERSITY",
                    "id": "6333fb3f-2838-4496-88bb-3e67160e2d4d",
                    "name": "DIVERSITY",
                    "state": "active",
                    "vocabulary_id": null
                },
                {
                    "display_name": "INSECTS",
                    "id": "727d285b-f4e8-4bc7-8e71-e63ed9252bf6",
                    "name": "INSECTS",
                    "state": "active",
                    "vocabulary_id": null
                },
                {
                    "display_name": "SDMS",
                    "id": "1feba39d-6f50-4300-b5ba-6adc00490ee0",
                    "name": "SDMS",
                    "state": "active",
                    "vocabulary_id": null
                },
                {
                    "display_name": "SNAILS",
                    "id": "92a892d7-13e7-43a7-98b2-339c6cb4b3c1",
                    "name": "SNAILS",
                    "state": "active",
                    "vocabulary_id": null
                },
                {
                    "display_name": "SPECIES ABUNDANCE DISTRIBUTIONS",
                    "id": "3d02d873-d64b-4164-b73a-d0123c3e7128",
                    "name": "SPECIES ABUNDANCE DISTRIBUTIONS",
                    "state": "active",
                    "vocabulary_id": null
                },
                {
                    "display_name": "SPECIES RICHNESS MODELS",
                    "id": "e61b9a3f-fa0d-4e90-ae32-279ad9ceda29",
                    "name": "SPECIES RICHNESS MODELS",
                    "state": "active",
                    "vocabulary_id": null
                },
                {
                    "display_name": "SPIDERS",
                    "id": "3762a970-222d-4cb3-ab58-bf79e164e5ed",
                    "name": "SPIDERS",
                    "state": "active",
                    "vocabulary_id": null
                },
                {
                    "display_name": "URBAN ECOLOGY",
                    "id": "68337815-ea2a-4f12-8c28-013f9971878e",
                    "name": "URBAN ECOLOGY",
                    "state": "active",
                    "vocabulary_id": null
                },
                {
                    "display_name": "URBANIZATION",
                    "id": "80141c83-e3e9-43b5-9c76-abad1e74852c",
                    "name": "URBANIZATION",
                    "state": "active",
                    "vocabulary_id": null
                }
            ],
            "groups": [],
            "relationships_as_subject": [],
            "relationships_as_object": []
        }"""
    ).strip()



"""Get EnviDat metadata records in various formats."""

import json
import logging
from typing import Literal, NoReturn, Union

from envidat.api.v1 import (
    get_metadata_list_with_resources,
    get_metadata_name_doi,
    get_package,
)
from envidat.converters.bibtex_converter import convert_bibtex
from envidat.converters.datacite_converter import convert_datacite
from envidat.converters.dcat_ap_converter import convert_dcat_ap
from envidat.converters.dif_converter import convert_dif
from envidat.converters.iso_converter import convert_iso
from envidat.converters.ris_converter import convert_ris
from envidat.converters.xml_converter import convert_xml

log = logging.getLogger(__name__)


def validate_json(json_data):
    """Test if JSON parses and is valid."""
    try:
        json.loads(json_data)
    except ValueError:
        return False
    return True


class Record:
    """Class manipulate an EnviDat record in various ways."""

    content = None

    def __init__(
        self,
        input_data: Union[str, dict],
        convert: Literal[
            "str", "xml", "iso", "bibtex", "dif", "datacite", "ris", "dcat-ap"
        ] = None,
    ) -> NoReturn:
        """
        Init the Record object.

        Only one argument should be passed for data format.

        Args:
            input_data [str, dict]: Data input, in JSON or dict form.
                Can also accept a package name to extract a record from the API.
            convert (str):
                Options: Convert the content immediately to specified type.
                    "str", "xml", "iso", "bibtex", "dif", "datacite", "ris", "dcat-ap"

        """
        if isinstance(input_data, dict):
            # Is dict
            log.debug("Dictionary input provided, reading as JSON")
            self.content = input_data

        elif isinstance(input_data, str):
            if validate_json(input_data):
                # Is JSON String, parse to JSON object/dict
                log.debug("Valid input JSON parsed")
                self.content = json.loads(input_data)
            else:
                # Get from API (JSON object/dict)
                log.debug("Attempting to get package JSON from API")
                self.content = dict(get_package(input_data))

        else:
            log.error("Input is not a valid type from (str,dict)")
            raise TypeError("Input must be of type string or dict")

        # Validate metadata record
        self.validate()

        if convert:
            mapping = {
                "json": self.to_json,
                "xml": self.to_xml,
                "iso": self.to_iso,
                "bibtex": self.to_bibtex,
                "dif": self.to_dif,
                "datacite": self.to_datacite,
                "ris": self.to_ris,
                "dcat-ap": self.to_dcat_ap,
            }
            if convert == "datacite":
                name_doi_map = get_metadata_name_doi()
                self.content = mapping[convert](name_doi_map)
            else:
                self.content = mapping[convert]()

    def get_content(self):
        """
        Get current content of Record.

        Returns:
            str: Metadata record, default dict format, else converted (JSON, XML, etc).
        """
        return self.content

    def validate(self) -> bool:
        """
        Validate metadata record.

        Returns:
            bool: True if valid, raises error if not.
        """
        metadata_keys = [
            "author",
            "author_email",
            "creator_user_id",
            "date",
            "doi",
            "funding",
            "id",
            "isopen",
            "language",
            "license_id",
            "license_title",
            "maintainer",
            "maintainer_email",
            "metadata_created",
            "metadata_modified",
            "name",
            "notes",
            "num_resources",
            "num_tags",
            "organization",
            "owner_org",
            "private",
            "publication",
            "publication_state",
            # "related_datasets", NOT ALWAYS PRESENT
            "related_publications",
            "resource_type",
            "resource_type_general",
            "spatial",
            "spatial_info",
            "state",
            "subtitle",
            "title",
            "type",
            "url",
            "version",
            "resources",
            "tags",
            "groups",
            "relationships_as_subject",
            "relationships_as_object",
        ]

        log.debug("Validating metadata record")
        if not isinstance(self.content, dict):
            log.error(f"Content is not a valid dictionary of metadata: {self.content}")
            raise ValueError("Content is not a valid dictionary of metadata.")

        missing_keys = list(set(metadata_keys) - set(self.content.keys()))
        if missing_keys:
            log.error(f"Metadata entry is missing fields: {missing_keys}")
            raise ValueError(
                "Content does not have all required fields for a metadata entry."
            )

        return True

    def to_json(self) -> str:
        """
        Convert content to JSON string.

        Returns:
            str: JSON string of metadata record.
        """
        return json.loads(self.content)

    def to_xml(self) -> str:
        """
        Convert content to XML format.

        Returns:
            str: XML formatted string of metadata record.
        """
        return convert_xml(self.content)

    def to_iso(self) -> str:
        """
        Convert content to ISO format.

        Returns:
            str: ISO formatted string of metadata record.
        """
        return convert_iso(self.content)

    def to_ris(self) -> str:
        """
        Convert content to RIS format.

        Returns:
            str: RIS formatted string of metadata record.
        """
        return convert_ris(self.content)

    def to_bibtex(self) -> str:
        """
        Convert content to BibTeX format.

        Returns:
            str: BibTeX formatted string of metadata record.
        """
        return convert_bibtex(self.content)

    def to_dif(self) -> str:
        """
        Convert content to GCMD DIF 10.2 format.

        Returns:
            str: GCMD DIF 10.2 formatted string of metadata record.
        """
        return convert_dif(self.content)

    def to_datacite(self, name_doi_map: dict) -> str:
        """
        Convert content to DataCite format.

        Args:
            name_doi_map (dict): Mapping of dataset name to DOI, format name:doi.

        Returns:
            str: DataCite formatted string of metadata record.
        """
        return convert_datacite(self.content, name_doi_map)

    def to_dcat_ap(self) -> str:
        """
        Convert content to DCAT-AP CH format.

        Returns:
            str: DCAT-AP CH formatted string of metadata record.
        """
        return convert_dcat_ap(self.content)


def get_all_metadata_record_list(
    convert: Literal[
        "str", "xml", "iso", "bibtex", "dif", "datacite", "ris", "dcat-ap"
    ] = None,
    content_only: bool = False,
) -> Union[list, str]:
    """
    Return all EnviDat metadata entries as Record objects.

    Defaults to standard Record, content in json format.

    Args:
        convert (str): Convert the content immediately to specified type.
            Options: "str", "xml", "iso", "bibtex", "dif", "datacite", "ris", "dcat-ap"
        content_only (bool): Extract content from Record objects.

    Returns:
        (list, string): Of Record entries for EnviDat metadata.
            Note: returns a string in the case of a single metadata entry
            and content_only is set to True.
    """
    metadata = get_metadata_list_with_resources()

    # DCAT-AP special case, return as single XML
    if convert == "dcat-ap":
        loaded_metadata = [Record(entry).content for entry in metadata]
        return convert_dcat_ap(loaded_metadata)

    # Only get metadata:doi mapping once
    if convert == "datacite":
        name_doi_map = get_metadata_name_doi()

    record_list = []

    for metadata_entry in metadata:
        record = Record(metadata_entry)

        if convert:
            mapping = {
                "json": record.to_json,
                "xml": record.to_xml,
                "iso": record.to_iso,
                "bibtex": record.to_bibtex,
                "dif": record.to_dif,
                "datacite": record.to_datacite,
                "ris": record.to_ris,
                # dcat-ap handled separately above
            }

            if convert == "datacite":
                record.content = mapping[convert](name_doi_map)
            else:
                record.content = mapping[convert]()

        if content_only:
            record_list.append(record.content)
        else:
            record_list.append(record)

    return record_list



"""Tests for metadata Record."""

from collections import Counter

import pytest
from xmltodict import parse

from envidat.metadata import Record, get_all_metadata_record_list


def test_record_init_from_json(example_ckan_json, metadata_keys):
    """Test init using JSON input."""
    record = Record(example_ckan_json)

    assert Counter(list(record.content.keys())) == Counter(metadata_keys)


def test_record_init_from_dict(example_ckan_dict, metadata_keys):
    """Test init using dict input."""
    record = Record(example_ckan_dict)

    assert Counter(list(record.content.keys())) == Counter(metadata_keys)


def test_record_init_from_api(example_ckan_json, metadata_keys):
    """Test init using call to API."""
    record_from_api = Record(
        "ecological-properties-of-urban-ecosystems-biodiversity-dataset-of-zurich"
    )

    assert Counter(list(record_from_api.content.keys())) == Counter(metadata_keys)

    record_from_json = Record(example_ckan_json)
    assert record_from_api.content == record_from_json.content


def test_record_init_with_validate(example_ckan_json, metadata_keys):
    """Test manual trigger of incorrect validation."""
    record = Record(example_ckan_json)

    is_valid = record.validate()
    assert is_valid

    record.content.pop("title")
    with pytest.raises(
        ValueError,
        match=("Content does not have all required fields for a metadata entry."),
    ):
        is_valid = record.validate()

    record.content = "not a valid json"
    with pytest.raises(
        ValueError, match="Content is not a valid dictionary of metadata."
    ):
        is_valid = record.validate()


def test_record_init_with_convert(example_ckan_json):
    """Test auto-conversion to another format during Record init."""
    record = Record(example_ckan_json)
    record_converted = Record(example_ckan_json, convert="xml")

    xml = record.to_xml()
    assert record_converted.content == xml


def test_get_all_metadata_record_list():
    """Test getting all metadata entries as list of Records."""
    metadata_records = get_all_metadata_record_list()

    assert len(metadata_records) > 500


def test_get_all_metadata_xml_list():
    """Test conversion of all Record contents to XML."""
    metadata_records = get_all_metadata_record_list(convert="xml")

    assert len(metadata_records) > 500
    assert type(metadata_records[0] is Record)


def test_get_all_metadata_iso_list_content_only():
    """Test converting and extracting contents only for each Record."""
    metadata_records = get_all_metadata_record_list(convert="iso", content_only=True)

    assert len(metadata_records) > 500
    assert type(metadata_records[0] is dict)


def test_get_all_metadata_datacite():
    """Test conversion of all to Datacite format, specific case."""
    metadata_records = get_all_metadata_record_list(convert="datacite")

    assert len(metadata_records) > 500


def test_get_single_dcat_ap_xml(example_ckan_json):
    """Test conversion of single Record to DCAT-AP format."""
    record = Record(example_ckan_json)
    record_converted = Record(example_ckan_json, convert="dcat-ap")

    xml = record.to_dcat_ap()
    assert record_converted.content == xml


def test_get_all_metadata_dcat_ap():
    """Test conversion of all to DCAT-AP format, specific case."""
    metadata_records = get_all_metadata_record_list(convert="dcat-ap")

    assert len(metadata_records) > 500


def test_get_all_metadata_dcat_ap_formatted_xml():
    """Test conversion of all to single DCAT-AP XML, specific case."""
    dcat_ap_xml = get_all_metadata_record_list(convert="dcat-ap", content_only=True)

    assert isinstance(dcat_ap_xml, str)

    dcat_dict = parse(dcat_ap_xml)
    assert list(dcat_dict.keys())[0] == "rdf:RDF"

    dataset_list = dcat_dict["rdf:RDF"]["dcat:Catalog"]["dcat:dataset"]
    assert len(dataset_list) > 500



"""DCAT-AP CH for OpenDataSwiss."""

import json
import logging
from collections import OrderedDict
from typing import Union

from dateutil.parser import parse as parse_date
from xmltodict import parse, unparse

log = logging.getLogger(__name__)


def convert_dcat_ap(metadata_records: Union[dict, list[dict]]) -> str:
    """
    Generate output string in DCAT-AP format.

    Accepts a single metadata entry, or list of entries.

    Note:
        Converter is only valid for the metadata schema for EnviDat.

    Args:
        metadata_records (dict, list[dict]):
            Either:
                - Individual EnviDat metadata entry dictionary.
                - List of EnviDat metadata record dictionaries.

    Returns:
        str: string in DCAT-AP CH XML format.
    """
    try:

        if isinstance(metadata_records, list):
            converted_records = []
            for record in metadata_records:
                converted_records.append(
                    dcat_ap_convert_dataset(
                        record,
                    )
                )
            return wrap_packages_dcat_ap_xml(converted_records)

        return wrap_packages_dcat_ap_xml(
            dcat_ap_convert_dataset(
                metadata_records,
                as_xml=True,
            )
        )
    except AttributeError as e:
        log.error(e)
        log.error("Cannot convert package to DCAT-AP format.")
        raise AttributeError("Failed to convert package to DCAT-AP format.")


def dcat_ap_convert_dataset(
    metadata_record: dict, as_xml: bool = False
) -> Union[str, dict]:
    """Generate output string in DCAT-AP format.

    Note:
        Converter is only valid for the metadata schema for EnviDat.

    Args:
        metadata_record (dict): Individual EnviDat metadata entry record dictionary.
        as_xml (bool): If set, unparse dictionary to XML string.

    Returns:
        (str, dict):
            Either:
                - XML string in DCAT-AP format, if as_xml=True.
                - Dictionary in DCAT-AP format, for further processing.

    """
    md_metadata_dict = OrderedDict()

    # Dataset URL
    package_name = metadata_record["name"]
    package_url = f"https://www.envidat.ch/#/metadata/{package_name}"
    md_metadata_dict["dcat:Dataset"] = {"@rdf:about": package_url}

    # identifier
    package_id = metadata_record["id"]
    md_metadata_dict["dcat:Dataset"]["dct:identifier"] = f"{package_id}@envidat"

    # title
    title = metadata_record["title"]
    md_metadata_dict["dcat:Dataset"]["dct:title"] = {
        "@xml:lang": "en",
        "#text": title,
    }

    # description
    description = clean_text(metadata_record.get("notes", ""))
    md_metadata_dict["dcat:Dataset"]["dct:description"] = {
        "@xml:lang": "en",
        "#text": description,
    }

    # issued
    creation_date = metadata_record.get("metadata_created")
    if creation_date:
        md_metadata_dict["dcat:Dataset"]["dct:issued"] = {
            "@rdf:datatype": "http://www.w3.org/2001/XMLSchema#dateTime",
            "#text": parse_date(creation_date).strftime("%Y-%m-%dT%H:%M:%SZ"),
        }

    # modified
    modification_date = metadata_record.get("metadata_modified", creation_date)
    if modification_date:
        md_metadata_dict["dcat:Dataset"]["dct:modified"] = {
            "@rdf:datatype": "http://www.w3.org/2001/XMLSchema#dateTime",
            "#text": parse_date(modification_date).strftime("%Y-%m-%dT%H:%M:%SZ"),
        }

    # publication (MANDATORY)
    md_metadata_dict["dcat:Dataset"]["dct:publisher"] = {
        "foaf:Organization": {
            "@rdf:about": "https://envidat.ch/#/about",
            "foaf:name": "EnviDat",
        }
    }

    # landing page
    md_metadata_dict["dcat:Dataset"]["dcat:landingPage"] = {
        "@rdf:resource": package_url
    }

    # contact point (MANDATORY)
    maintainer = json.loads(metadata_record.get("maintainer", "{}"))
    maintainer_name = ""

    if maintainer.get("given_name"):
        maintainer_name += maintainer["given_name"].strip() + " "

    maintainer_name += maintainer["name"]
    maintainer_email = "mailto:" + maintainer["email"]
    individual_contact_point = {
        "vcard:Individual": {
            "vcard:fn": maintainer_name,
            "vcard:hasEmail": {"@rdf:resource": maintainer_email},
        }
    }

    if maintainer_email == "mailto:envidat@wsl.ch":
        md_metadata_dict["dcat:Dataset"]["dcat:contactPoint"] = [
            individual_contact_point
        ]
    else:
        organization_contact_point = {
            "vcard:Organization": {
                "vcard:fn": "EnviDat Support",
                "vcard:hasEmail": {"@rdf:resource": "mailto:envidat@wsl.ch"},
            }
        }
        md_metadata_dict["dcat:Dataset"]["dcat:contactPoint"] = [
            individual_contact_point,
            organization_contact_point,
        ]

    # theme (MANDATORY)
    md_metadata_dict["dcat:Dataset"]["dcat:theme"] = {
        "@rdf:resource": "http://opendata.swiss/themes/education"
    }

    # language
    md_metadata_dict["dcat:Dataset"]["dct:language"] = {"#text": "en"}

    # keyword
    keywords_list = []
    keywords = get_keywords(metadata_record)
    for keyword in keywords:
        keywords_list += [{"@xml:lang": "en", "#text": keyword}]
    md_metadata_dict["dcat:Dataset"]["dcat:keyword"] = keywords_list

    # Distribution - iterate through package resources and obtain
    # package license (MANDATORY)
    # Call get_distribution_list(metadata_record) to get distibution list
    md_metadata_dict["dcat:Dataset"]["dcat:distribution"] = get_distribution_list(
        metadata_record, package_name
    )

    if as_xml:
        return unparse(md_metadata_dict, short_empty_elements=True, pretty=True)
    return md_metadata_dict


def clean_text(text: str) -> str:
    """Text cleaned of hashes and with modified characters."""
    cleaned_text = (
        text.replace("###", "")
        .replace("##", "")
        .replace(" #", " ")
        .replace("# ", " ")
        .replace("__", "")
        .replace("  ", " ")
        .replace("\r", "\n")
        .replace("\n\n", "\n")
    )
    return cleaned_text


def get_keywords(metadata_record: dict) -> list:
    """Keywords from tags in package (metadata record)."""
    keywords = []
    for tag in metadata_record.get("tags", []):
        name = tag.get("display_name", "").upper()
        keywords += [name]
    return keywords


def get_distribution_list(metadata_record: dict, package_name: str) -> list:
    """Return distribution_list created from package resources list and licence_id."""
    distribution_list = []

    dataset_license = metadata_record.get("license_id", "odc-odbl")

    license_mapping = {
        "wsl-data": (
            "NonCommercialWithPermission-CommercialWithPermission-ReferenceRequired"
        ),
        "odc-odbl": "NonCommercialAllowed-CommercialAllowed-ReferenceRequired",
        "cc-by": "NonCommercialAllowed-CommercialAllowed-ReferenceRequired",
        "cc-by-sa": "NonCommercialAllowed-CommercialAllowed-ReferenceRequired",
        "cc-zero": "NonCommercialAllowed-CommercialAllowed-ReferenceNotRequired",
        "CC0-1.0": "NonCommercialAllowed-CommercialAllowed-ReferenceNotRequired",
    }
    resource_license = license_mapping.get(
        dataset_license,
        "NonCommercialWithPermission-CommercialWithPermission-ReferenceRequired",
    )

    for resource in metadata_record.get("resources", []):

        resource_id = resource.get("id")
        resource_name = resource.get("name", resource_id)
        resource_notes = clean_text(resource.get("description", "No description"))
        resource_page_url = (
            f"https://www.envidat.ch/dataset/{package_name}/resource/"
            + resource.get("id", "")
        )
        resource_url = resource.get("url")

        resource_creation = parse_date(resource["created"]).strftime(
            "%Y-%m-%dT%H:%M:%SZ"
        )
        resource_modification = resource_creation
        if resource.get("last_modified", resource.get("metadata_modified", "")):
            resource_modification = parse_date(
                resource.get("last_modified", resource.get("metadata_modified", ""))
            ).strftime("%Y-%m-%dT%H:%M:%SZ")

        resource_size = resource.get("size", False)

        if not resource_size:
            resource_size = 0
            try:
                if len(resource.get("resource_size", "")) > 0:
                    resource_size_obj = json.loads(
                        resource.get("resource_size", "{'size_value': '0'}")
                    )
                    sizes_dict = {
                        "KB": 1024,
                        "MB": 1048576,
                        "GB": 1073741824,
                        "TB": 1099511627776,
                    }
                    resource_size_str = resource_size_obj.get("size_value", "")
                    if len(resource_size_str) > 0:
                        resource_size = (
                            float(resource_size_obj.get("size_value", "0"))
                            * sizes_dict[
                                resource_size_obj.get("size_units", "KB").upper()
                            ]
                        )
            except Exception as e:
                log.error(
                    "resource {} unparseable resource_size: {}".format(
                        resource_url, resource.get("resource_size")
                    )
                )
                log.error(e)
                resource_size = 0

        resource_mimetype = resource.get("mimetype", "")
        if not resource_mimetype or len(resource_mimetype) == 0:
            resource_mimetype = resource.get("mimetype_inner")

        resource_format = resource.get("format")

        distribution = {
            "dcat:Distribution": {
                "@rdf:about": resource_page_url,
                "dct:identifier": metadata_record["name"] + "." + resource_id,
                "dct:title": {"@xml:lang": "en", "#text": resource_name},
                "dct:description": {"@xml:lang": "en", "#text": resource_notes},
                "dct:issued": {
                    "@rdf:datatype": "http://www.w3.org/2001/XMLSchema#dateTime",
                    "#text": resource_creation,
                },
                "dct:modified": {
                    "@rdf:datatype": "http://www.w3.org/2001/XMLSchema#dateTime",
                    "#text": resource_modification,
                },
                "dct:language": "en",
                "dcat:accessURL": {
                    "@rdf:resource": resource_page_url,
                },
                "dct:rights": resource_license,
                "dcat:byteSize": resource_size,
            }
        }
        # mediaType
        if resource_mimetype:
            distribution["dcat:Distribution"]["dcat:mediaType"] = resource_mimetype

        # format
        if resource_format:
            distribution["dcat:Distribution"]["dct:format"] = resource_format

        distribution_list += [distribution]

    return distribution_list


def wrap_packages_dcat_ap_xml(dcat_xml_packages: list) -> str:
    """
    Add required DCAT-AP catalog XML tags for full DCAT-AP XML.

    Args:
        dcat_xml_packages (list[str,dict]): All DCAT-AP formatted packages to include.
            In string XML or dictionary format.

    Note: this is a required final step for producing a DCAT-AP CH format XML.
    """
    catalog_dict = OrderedDict()

    # header
    catalog_dict["@xmlns:dct"] = "http://purl.org/dc/terms/"
    catalog_dict["@xmlns:dc"] = "http://purl.org/dc/elements/1.1/"
    catalog_dict["@xmlns:dcat"] = "http://www.w3.org/ns/dcat#"
    catalog_dict["@xmlns:foaf"] = "http://xmlns.com/foaf/0.1/"
    catalog_dict["@xmlns:xsd"] = "http://www.w3.org/2001/XMLSchema#"
    catalog_dict["@xmlns:rdfs"] = "http://www.w3.org/2000/01/rdf-schema#"
    catalog_dict["@xmlns:rdf"] = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
    catalog_dict["@xmlns:vcard"] = "http://www.w3.org/2006/vcard/ns#"
    catalog_dict["@xmlns:odrs"] = "http://schema.theodi.org/odrs#"
    catalog_dict["@xmlns:schema"] = "http://schema.org/"

    if isinstance(dcat_xml_packages, dict):
        packages_dict_list = dcat_xml_packages
    elif isinstance(dcat_xml_packages, str):
        packages_dict_list = parse(dcat_xml_packages)
    elif isinstance(dcat_xml_packages, list):
        if isinstance(dcat_xml_packages[0], dict):
            packages_dict_list = dcat_xml_packages
        elif isinstance(dcat_xml_packages[0], str):
            packages_dict_list = [parse(package) for package in dcat_xml_packages]
    else:
        log.error("Packages in incorrect format. Must be string XML or dict.")
        raise ValueError("Packages in incorrect format. Must be string XML or dict.")
    catalog_dict["dcat:Catalog"] = {"dcat:dataset": packages_dict_list}

    # Assign dcat_catalog_dict dictionary for root element in XML file
    dcat_catalog_dict = OrderedDict()
    dcat_catalog_dict["rdf:RDF"] = catalog_dict

    return unparse(dcat_catalog_dict, short_empty_elements=True, pretty=True)
